#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/times.h>
#include <asm/param.h>  /* HZ */
#include <openssl/bn.h>
#include <openssl/evp.h>

char *Progname;
#define USAGE() fprintf(stderr, "usage: %s [-d dictfile] verifier\n", Progname)

int Words;

void doit(char *, char *);
void parse_verifier(FILE *infile, char **name,
		    BIGNUM *p, BIGNUM *g, BIGNUM *v, char **salt, u_int *slen);
void srp_x_calc(const char *salt, int slen, const char *username,
		const char *passphrase, BIGNUM *x);
int srp_tfmt2bn(BIGNUM *dst, const char *src);

/* Dictionary attack on an SRP verifier. */

int main(int argc, char **argv)
{
	int c;
	char *dictfile;
	double wps;
	struct tms t0, t1;

	Progname = *argv;

	/* Parse option arguments. */

	dictfile = "/usr/dict/words";
	while ((c = getopt(argc, argv, "d:")) != EOF) {
		switch (c) {
		case 'd':
			dictfile = optarg;
			break;

		default:
			USAGE();
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 1) {
		USAGE();
		exit(1);
	}

	Words = 0;
	times(&t0);
	doit(argv[0], dictfile);
	times(&t1);

	wps = Words / ((double)(t1.tms_utime - t0.tms_utime) / HZ);
	printf("%lf wps\n", wps);

	return 0;
}

void doit(char *vfile, char *dictfile)
{
	int i, slen;
	char buf[100];
	char *username, *salt;
	FILE *infile;
	BIGNUM *p, *g, *s, *x, *v, *v0;
	BN_CTX *ctx, *ctx2;

	/* Allocate BIGNUMs and stuff. */

	ctx = BN_CTX_new();
	ctx2 = BN_CTX_new();
	BN_CTX_start(ctx);
	p = BN_CTX_get(ctx);
	g = BN_CTX_get(ctx);
	s = BN_CTX_get(ctx);
	x = BN_CTX_get(ctx);
	v = BN_CTX_get(ctx);
	v0 = BN_CTX_get(ctx);

	/* Get the entry from the verifier file. */

	infile = fopen(vfile, "r");
	if (infile == NULL) {
		fprintf(stderr, "can't open '%s'\n", vfile);
		exit(1);
	}

	parse_verifier(infile, &username, p, g, v0, &salt, &slen);

	fclose(infile);

	/* Open the dict file. */

	infile = fopen(dictfile, "r");
	if (infile == NULL) {
		fprintf(stderr, "can't open '%s'\n", dictfile);
		exit(1);
	}

	/* For each line in the file, make a verifier and compare it with
	the one from the file. */

	while (fgets(buf, sizeof(buf), infile)) {
		buf[strlen(buf) - 1] = '\0';

		/* Calculate x = HASH(salt | HASH(username | passphrase)). */

		srp_x_calc(salt, slen, username, buf, x);

		/* Calculate v = g^x mod p. */

		BN_mod_exp(v, g, x, p, ctx2);
		Words++;

		/* Compare. */

		if (BN_ucmp(v, v0) == 0) {
			printf("%s\n", buf);
			return;
		}
	}
	fclose(infile);
}

/* Return the next verifier line from the file. */

void parse_verifier(FILE *infile, char **name,
		    BIGNUM *p, BIGNUM *g, BIGNUM *v, char **salt, u_int *slen)
{
	char *s, *uptr, *vptr, *sptr, *pptr, *gptr;
	char buf[8 * 1024];
	int c, n;

	/* Get the next line.  It's probably long, and the 8K buffer is
	enough to read some very very large numbers.  Normal use with 1024
	bit numbers is less than 1K. */

	s = buf;
	n = 0;
	while ((c = getc(infile)) != EOF && c != '\n') {
		*s++ = c;
		n++;
		if (n == sizeof(buf))
			goto parse_err;
	}
	*s = '\0';

	if (n == 0 && c == EOF) {
		fprintf(stderr, "EOF\n");
		exit(0);
	}

	/* Carve up the buffer. u:v:s::p:g */

	s = buf;
	if ((uptr = strsep(&s, ":")) == NULL)
		goto parse_err;
	if ((vptr = strsep(&s, ":")) == NULL)
		goto parse_err;
	if ((sptr = strsep(&s, ":")) == NULL)
		goto parse_err;
	if (*s++ != ':')
		goto parse_err;
	if ((pptr = strsep(&s, ":")) == NULL)
		goto parse_err;
	if ((gptr = strsep(&s, ":")) == NULL)
		goto parse_err;
	if (s != NULL)
		goto parse_err;

	/* Convert from base64. */

	*name = strdup(uptr);

	srp_tfmt2bn(v, vptr);

	*slen = srp_tfmt2bin(sptr, sptr);
	*salt = malloc(*slen);
	memcpy(*salt, sptr, *slen);

	srp_tfmt2bn(p, pptr);

	srp_tfmt2bn(g, gptr);

	return;

parse_err:
	fprintf(stderr, "verifier file parse error\n");
	exit(1);
}

/*
 * Calculate x = HASH(salt | HASH(username | passphrase)).  This
 * calculation is used both in verifier generation and in the client.
 * Sets x, which must be allocated.
 */
void srp_x_calc(const char *salt, int slen, const char *username,
		const char *passphrase, BIGNUM *x)
{
	int dlen;
	u_char digest[EVP_MAX_MD_SIZE];
	EVP_MD *evp_md = EVP_sha1();
	EVP_MD_CTX md;

	EVP_DigestInit(&md, evp_md);
	EVP_DigestUpdate(&md, username, strlen(username));
	EVP_DigestUpdate(&md, ":", 1);
	EVP_DigestUpdate(&md, passphrase, strlen(passphrase));
	EVP_DigestFinal(&md, digest, &dlen);

	EVP_DigestInit(&md, evp_md);
	EVP_DigestUpdate(&md, salt, slen);
	EVP_DigestUpdate(&md, digest, dlen);
	EVP_DigestFinal(&md, digest, &dlen);

	BN_bin2bn(digest, dlen, x);
}

static char b64table[] =
    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz./";

/*
 * Convert a libsrp base64 string into a BIGNUM, which must be
 * preallocated.  Returns 1 for success, 0 for error.
 */
int srp_tfmt2bn(BIGNUM *dst, const char *src)
{
	char *a;
	int i, ret;
	int srp_tfmt2bin(char *dst, const char *src);

	i = strlen(src) + 1;
	a = malloc(i);

	i = srp_tfmt2bin(a, src);
	ret = 1;
	if (BN_bin2bn(a, i, dst) == NULL)
		ret = 0;

	free(a);
	return ret;
}

/*
 * Convert a libsrp base64 string into raw byte array representation.
 * dst must already be allocated, and must be large enough.  Returns the
 * length of the converted src.
 */
int srp_tfmt2bin(char *dst, const char *src)
{
	char *loc;
	u_char *a;
	int i, j, size;

	while (*src && (*src == ' ' || *src == '\t' || *src == '\n'))
		++src;
	size = strlen(src);
	if (size == 0)
		return size;

	a = malloc((size + 1) * sizeof(u_char));
	i = 0;
	while (i < size) {
		loc = strchr(b64table, src[i]);
		if (loc == (char *) 0)
			break;
		else
			a[i] = loc - b64table;
		++i;
	}
	size = i;

	i = size - 1;
	j = size;
	for (;;) {
		a[j] = a[i];
		if (--i < 0)
			break;
		a[j] |= (a[i] & 3) << 6;
		--j;
		a[j] = (u_char) ((a[i] & 0x3c) >> 2);
		if (--i < 0)
			break;
		a[j] |= (a[i] & 0xf) << 4;
		--j;
		a[j] = (u_char) ((a[i] & 0x30) >> 4);
		if (--i < 0)
			break;
		a[j] |= (a[i] << 2);

		a[--j] = 0;
		if (--i < 0)
			break;
	}

	while (a[j] == 0 && j <= size)
		++j;

	memcpy(dst, a + j, size - j + 1);
	free(a);
	return size - j + 1;
}
